from collections.abc import Sequence
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Self

from tqdm import tqdm

from karaokatalog.Song import Song


@dataclass(frozen=True)
class Library:
    songs: Sequence[Song]
    unparseable_song_txts: Sequence[Path]

    @classmethod
    def from_dir(cls, library_dir: Path) -> Self:
        """
        Load a library from a directory by recursively finding all *.txt files in there and
        treating them as UltraStar DX Song TXT files.
        Any *.txt file that cannot be parsed is listed as unparseable Song TXT.
        """
        maybe_songs_by_path = {
            song_txt: Song.from_song_txt(song_txt)
            for song_txt in tqdm(library_dir.rglob("*.txt"), unit=" songs")
        }

        songs = tuple(song for song in maybe_songs_by_path.values() if song)
        unparseable_song_txts = tuple(
            song_txt for song_txt, song in maybe_songs_by_path.items() if not song
        )

        return cls(songs, unparseable_song_txts)

    @cached_property
    def songs_in_unique_dirs(self) -> Sequence[Song]:
        """
        Sometimes, more than one song can share the same base directory in a library (this usually happens when there are variants
        of the song which share the same audio file).

        For deduplication, it is helpful to assume every song is in its own directory. This returns a sequence of songs filtered so
        that there are no two songs sharing the same base dir. This filtering is done arbitrarily.
        """
        unique_dirs: set[Path] = set()
        songs_in_unique_dirs: list[Song] = []

        for song in self.songs:
            if song.dir not in unique_dirs:
                songs_in_unique_dirs.append(song)
                unique_dirs.add(song.dir)

        return tuple(songs_in_unique_dirs)
